@@ -26,10 +26,10 @@ were all consulted.
 import struct
 import sys
 
-from mutagen import FileType, Metadata, StreamInfo, PaddingInfo
+from mutagen import FileType, Tags, StreamInfo, PaddingInfo
 from mutagen._constants import GENRES
-from mutagen._util import (cdata, insert_bytes, DictProxy, MutagenError,
-                           hashable, enum, get_size, resize_bytes)
+from mutagen._util import cdata, insert_bytes, DictProxy, MutagenError, \
+    hashable, enum, get_size, resize_bytes, loadfile, convert_error
 from mutagen._compat import (reraise, PY2, string_types, text_type, chr_,
                              iteritems, PY3, cBytesIO, izip, xrange)
 from ._atom import Atoms, Atom, AtomError
@@ -37,7 +37,7 @@ from ._util import parse_full_atom
 from ._as_entry import AudioSampleEntry, ASEntryError
 
 
-class error(IOError, MutagenError):
+class error(MutagenError):
     pass
 
 
@@ -58,7 +58,7 @@ __all__ = ['MP4', 'Open', 'delete', 'MP4Cover', 'MP4FreeForm', 'AtomDataType']
 
 @enum
 class AtomDataType(object):
-    """Enum for `dataformat` attribute of MP4FreeForm.
+    """Enum for ``dataformat`` attribute of MP4FreeForm.
 
     .. versionadded:: 1.25
     """
@@ -132,8 +132,8 @@ class MP4Cover(bytes):
     """A cover artwork.
 
     Attributes:
-
-    * imageformat -- format of the image (either FORMAT_JPEG or FORMAT_PNG)
+        imageformat (`AtomDataType`): format of the image
+            (either FORMAT_JPEG or FORMAT_PNG)
     """
 
     FORMAT_JPEG = AtomDataType.JPEG
@@ -168,8 +168,7 @@ class MP4FreeForm(bytes):
     """A freeform value.
 
     Attributes:
-
-    * dataformat -- format of the data (see AtomDataType)
+        dataformat (`AtomDataType`): format of the data (see AtomDataType)
     """
 
     FORMAT_DATA = AtomDataType.IMPLICIT  # deprecated
@@ -237,8 +236,25 @@ def _find_padding(atom_path):
         pass
 
 
-class MP4Tags(DictProxy, Metadata):
-    r"""Dictionary containing Apple iTunes metadata list key/values.
+def _item_sort_key(key, value):
+    # iTunes always writes the tags in order of "relevance", try
+    # to copy it as closely as possible.
+    order = ["\xa9nam", "\xa9ART", "\xa9wrt", "\xa9alb",
+             "\xa9gen", "gnre", "trkn", "disk",
+             "\xa9day", "cpil", "pgap", "pcst", "tmpo",
+             "\xa9too", "----", "covr", "\xa9lyr"]
+    order = dict(izip(order, xrange(len(order))))
+    last = len(order)
+    # If there's no key-based way to distinguish, order by length.
+    # If there's still no way, go by string comparison on the
+    # values, so we at least have something determinstic.
+    return (order.get(key[:4], last), len(repr(value)), repr(value))
+
+
+class MP4Tags(DictProxy, Tags):
+    r"""MP4Tags()
+
+    Dictionary containing Apple iTunes metadata list key/values.
 
     Keys are four byte identifiers, except for freeform ('----')
     keys. Values are usually unicode strings, but some atoms have a
@@ -305,7 +321,9 @@ class MP4Tags(DictProxy, Metadata):
 
     def __init__(self, *args, **kwargs):
         self._failed_atoms = {}
-        super(MP4Tags, self).__init__(*args, **kwargs)
+        super(MP4Tags, self).__init__()
+        if args or kwargs:
+            self.load(*args, **kwargs)
 
     def load(self, atoms, fileobj):
         try:
@@ -337,42 +355,31 @@ class MP4Tags(DictProxy, Metadata):
     def __setitem__(self, key, value):
         if not isinstance(key, str):
             raise TypeError("key has to be str")
+        self._render(key, value)
         super(MP4Tags, self).__setitem__(key, value)
 
     @classmethod
     def _can_load(cls, atoms):
         return b"moov.udta.meta.ilst" in atoms
 
-    @staticmethod
-    def _key_sort(item):
-        (key, v) = item
-        # iTunes always writes the tags in order of "relevance", try
-        # to copy it as closely as possible.
-        order = ["\xa9nam", "\xa9ART", "\xa9wrt", "\xa9alb",
-                 "\xa9gen", "gnre", "trkn", "disk",
-                 "\xa9day", "cpil", "pgap", "pcst", "tmpo",
-                 "\xa9too", "----", "covr", "\xa9lyr"]
-        order = dict(izip(order, xrange(len(order))))
-        last = len(order)
-        # If there's no key-based way to distinguish, order by length.
-        # If there's still no way, go by string comparison on the
-        # values, so we at least have something determinstic.
-        return (order.get(key[:4], last), len(repr(v)), repr(v))
+    def _render(self, key, value):
+        atom_name = _key2name(key)[:4]
+        if atom_name in self.__atoms:
+            render_func = self.__atoms[atom_name][1]
+        else:
+            render_func = type(self).__render_text
 
-    def save(self, filename, padding=None):
-        """Save the metadata to the given filename."""
+        return render_func(self, key, value)
+
+    @convert_error(IOError, error)
+    @loadfile(writable=True)
+    def save(self, filething, padding=None):
 
         values = []
-        items = sorted(self.items(), key=self._key_sort)
+        items = sorted(self.items(), key=lambda kv: _item_sort_key(*kv))
         for key, value in items:
-            atom_name = _key2name(key)[:4]
-            if atom_name in self.__atoms:
-                render_func = self.__atoms[atom_name][1]
-            else:
-                render_func = type(self).__render_text
-
             try:
-                values.append(render_func(self, key, value))
+                values.append(self._render(key, value))
             except (TypeError, ValueError) as s:
                 reraise(MP4MetadataValueError, s, sys.exc_info()[2])
 
@@ -390,13 +397,12 @@ class MP4Tags(DictProxy, Metadata):
         data = Atom.render(b"ilst", b"".join(values))
 
         # Find the old atoms.
-        with open(filename, "rb+") as fileobj:
-            try:
-                atoms = Atoms(fileobj)
-            except AtomError as err:
-                reraise(error, err, sys.exc_info()[2])
+        try:
+            atoms = Atoms(filething.fileobj)
+        except AtomError as err:
+            reraise(error, err, sys.exc_info()[2])
 
-            self.__save(fileobj, atoms, data, padding)
+        self.__save(filething.fileobj, atoms, data, padding)
 
     def __save(self, fileobj, atoms, data, padding):
         try:
@@ -406,11 +412,6 @@ class MP4Tags(DictProxy, Metadata):
         else:
             self.__save_existing(fileobj, atoms, path, data, padding)
 
-    def __pad_ilst(self, data, length=None):
-        if length is None:
-            length = ((len(data) + 1023) & ~1023) - len(data)
-        return Atom.render(b"free", b"\x00" * length)
-
     def __save_new(self, fileobj, atoms, ilst_data, padding_func):
         hdlr = Atom.render(b"hdlr", b"\x00" * 8 + b"mdirappl" + b"\x00" * 9)
         meta_data = b"\x00\x00\x00\x00" + hdlr + ilst_data
@@ -621,7 +622,11 @@ class MP4Tags(DictProxy, Metadata):
 
     def __render_pair(self, key, value):
         data = []
-        for (track, total) in value:
+        for v in value:
+            try:
+                track, total = v
+            except TypeError:
+                raise ValueError
             if 0 <= track < 1 << 16 and 0 <= total < 1 << 16:
                 data.append(struct.pack(">4H", 0, track, total, 0))
             else:
@@ -761,7 +766,10 @@ class MP4Tags(DictProxy, Metadata):
             if not isinstance(v, text_type):
                 if PY3:
                     raise TypeError("%r not str" % v)
-                v = v.decode("utf-8")
+                try:
+                    v = v.decode("utf-8")
+                except (AttributeError, UnicodeDecodeError) as e:
+                    raise TypeError(e)
             encoded.append(v.encode("utf-8"))
 
         return self.__render_data(key, 0, flags, encoded)
@@ -819,25 +827,27 @@ class MP4Tags(DictProxy, Metadata):
 
 
 class MP4Info(StreamInfo):
-    """MPEG-4 stream information.
+    """MP4Info()
+
+    MPEG-4 stream information.
 
     Attributes:
+        bitrate (`int`): bitrate in bits per second, as an int
+        length (`float`): file length in seconds, as a float
+        channels (`int`): number of audio channels
+        sample_rate (`int`): audio sampling rate in Hz
+        bits_per_sample (`int`): bits per sample
+        codec (`mutagen.text`):
+            * if starting with ``"mp4a"`` uses an mp4a audio codec
+              (see the codec parameter in rfc6381 for details e.g.
+              ``"mp4a.40.2"``)
+            * for everything else see a list of possible values at
+              http://www.mp4ra.org/codecs.html
 
-    * bitrate -- bitrate in bits per second, as an int
-    * length -- file length in seconds, as a float
-    * channels -- number of audio channels
-    * sample_rate -- audio sampling rate in Hz
-    * bits_per_sample -- bits per sample
-    * codec (string):
-        * if starting with ``"mp4a"`` uses an mp4a audio codec
-          (see the codec parameter in rfc6381 for details e.g. ``"mp4a.40.2"``)
-        * for everything else see a list of possible values at
-          http://www.mp4ra.org/codecs.html
-
-        e.g. ``"mp4a"``, ``"alac"``, ``"mp4a.40.2"``, ``"ac-3"`` etc.
-    * codec_description (string):
-        Name of the codec used (ALAC, AAC LC, AC-3...). Values might change in
-        the future, use for display purposes only.
+            e.g. ``"mp4a"``, ``"alac"``, ``"mp4a.40.2"``, ``"ac-3"`` etc.
+        codec_description (`mutagen.text`):
+            Name of the codec used (ALAC, AAC LC, AC-3...). Values might
+            change in the future, use for display purposes only.
     """
 
     bitrate = 0
@@ -847,6 +857,7 @@ class MP4Info(StreamInfo):
     codec = u""
     codec_name = u""
 
+    @convert_error(IOError, MP4StreamInfoError)
     def __init__(self, atoms, fileobj):
         try:
             moov = atoms[b"moov"]
@@ -949,46 +960,58 @@ class MP4Info(StreamInfo):
 
 
 class MP4(FileType):
-    """An MPEG-4 audio file, probably containing AAC.
+    """MP4(filething)
+
+    An MPEG-4 audio file, probably containing AAC.
 
     If more than one track is present in the file, the first is used.
     Only audio ('soun') tracks will be read.
 
-    :ivar info: :class:`MP4Info`
-    :ivar tags: :class:`MP4Tags`
+    Arguments:
+        filething (filething)
+
+    Attributes:
+        info (`MP4Info`)
+        tags (`MP4Tags`)
     """
 
     MP4Tags = MP4Tags
 
     _mimes = ["audio/mp4", "audio/x-m4a", "audio/mpeg4", "audio/aac"]
 
-    def load(self, filename):
-        self.filename = filename
-        with open(filename, "rb") as fileobj:
-            try:
-                atoms = Atoms(fileobj)
-            except AtomError as err:
-                reraise(error, err, sys.exc_info()[2])
+    @loadfile()
+    def load(self, filething):
+        fileobj = filething.fileobj
 
+        try:
+            atoms = Atoms(fileobj)
+        except AtomError as err:
+            reraise(error, err, sys.exc_info()[2])
+
+        try:
+            self.info = MP4Info(atoms, fileobj)
+        except error:
+            raise
+        except Exception as err:
+            reraise(MP4StreamInfoError, err, sys.exc_info()[2])
+
+        if not MP4Tags._can_load(atoms):
+            self.tags = None
+            self._padding = 0
+        else:
             try:
-                self.info = MP4Info(atoms, fileobj)
+                self.tags = self.MP4Tags(atoms, fileobj)
             except error:
                 raise
             except Exception as err:
-                reraise(MP4StreamInfoError, err, sys.exc_info()[2])
-
-            if not MP4Tags._can_load(atoms):
-                self.tags = None
-                self._padding = 0
+                reraise(MP4MetadataError, err, sys.exc_info()[2])
             else:
-                try:
-                    self.tags = self.MP4Tags(atoms, fileobj)
-                except error:
-                    raise
-                except Exception as err:
-                    reraise(MP4MetadataError, err, sys.exc_info()[2])
-                else:
-                    self._padding = self.tags._padding
+                self._padding = self.tags._padding
+
+    def save(self, *args, **kwargs):
+        """save(filething=None, padding=None)"""
+
+        super(MP4, self).save(*args, **kwargs)
 
     def add_tags(self):
         if self.tags is None:
@@ -1004,7 +1027,19 @@ class MP4(FileType):
 Open = MP4
 
 
-def delete(filename):
-    """Remove tags from a file."""
+@convert_error(IOError, error)
+@loadfile(method=False, writable=True)
+def delete(filething):
+    """ delete(filething)
 
-    MP4(filename).delete()
+    Arguments:
+        filething (filething)
+    Raises:
+        mutagen.MutagenError
+
+    Remove tags from a file.
+    """
+
+    t = MP4(filething)
+    filething.fileobj.seek(0)
+    t.delete(filething)
